'' This object provides declarations and routines to manipulate
'' named object code spaces in I2C EEPROMS.  It also handles a
'' persistent storage area kept in the last long of the boot
'' EEPROM area.  This is overwritten when a new program is
'' downloaded to the EEPROM by the IDE, but is relocated to the
'' next "page" of the boot EEPROM when this is available.  This
'' area keeps a map of which I/O pins are used for I2C busses.
'' If present, the system automatically probes those busses for
'' EEPROMs during initialization.  There is a copy of the bitmap
'' stored in every code space when it is either named or deleted,
'' but the only copy referenced is the one in the second or first
'' code space of the boot EEPROM.

'' Named object code spaces are currently always 32Kx8 with the
'' name of the program stored in the last 4 longs (16 bytes) of
'' the 32Kx8 "page".  The last long contains system information
'' including a checksum so the file name is actually 12 chars.

'' These routines should be able to support another storage scheme,
'' but others may require some form of space compaction which can
'' be slow with a limited speed serial bus.

CON
  busBitmap   = $8000 - 2         ' Bitmap of pins with I2C busses
  checkMap    = $8000 - 3         ' Checksum for busBitmap
  checkSum    = $8000 - 4         ' Checksum for filename
  filename    = $8000 - 16        ' 12 character filename
  
  eepromDev   = %1010_000_0       ' EEPROM device address

  bootName    = $544F4F42         ' BOOT
  blankName   = $20202020         ' blanks

  opt_text    = %0000_0000_0001   ' file is text (not binary)

OBJ
  ldr   : "OS_loader"             ' System definitions
  data  : "OS_i2cObject"          ' EEPROM access via I2C
  vga   : "OS_vgaHiresText"       ' Hires VGA driver modified for OS use
  
PUB updateBitmapEEPROMs(addI2C) | i, t, m
'' This routine scans all the I2C busses identified in the stored
'' bitmap and fills in the EEPROM table based on what's present.
'' Any I2C busses that do not have EEPROMs are removed from the bitmap.
'' The parameter is either -1 or the number of a pin pair to try adding.
  result := getBusBitmap
  if addI2C <> -1
    result |= |<(addI2C >> 1)
  repeat i from 0 to (ldr#maxPinPairs << 1) - 2 step 2
    m := |<(i >> 1)
    if (result & m) <> 0      
      if (t := probeEEPROMs(i)) == 0
        if i == ldr#i2cBootSCL
          abort result     ' Boot EEPROM not present
        result &= !m
      word[long[ldr#systemPtr]+ldr#sys_eeproms+i] := t
  updateBitmap(result)

PRI getBusBitmap | i, j, mem, chkM, c, bitmap
'' This routine checks the first 32K of the boot EEPROM for a valid
'' I2C pin bitmap.  If there's more than 32K, the second 32K "page"
'' is checked first.  If there's a valid bitmap, this is returned.
'' Otherwise, a default bitmap showing only the boot EEPROM is used.
'' If the boot EEPROM is not accessable, the bitmap will be zero.
'' The resulting bitmap is returned.
  data.init(ldr#i2cBootSDA,ldr#i2cBootSCL)
  if data.testBus == data#_None
    if data.devicePresent(eepromDev)
      mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
      repeat i from 1 to 0
        if (i == 0) or (mem & |<i) <> 0
          bitmap~
          chkM := data.readLocation(eepromDev,(i<<15)+checkMap,16,8)
          repeat j from 0 to 1
            c := data.readLocation(eepromDev,(i<<15)+busBitmap+j,16,8)
            bitmap |= c << (8 * j)
            chkM += c
          if (chkM & $FF) == $FF
            return bitmap
      return |<(ldr#i2cBootSCL >> 1)

PRI updateBitmap(bitmap) | i, j, mem, chkM, c
'' This routine updates the I2C bus bitmap stored in the boot
'' EEPROM.  If there's a second "page", the bitmap is recorded
'' there.  Otherwise, it's saved in the first (or only) 32K.
  data.init(ldr#i2cBootSDA,ldr#i2cBootSCL)
  mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
  repeat i from 1 to 0
    if |<i & mem
      chkM := $FF
      repeat j from 0 to 1
        c := bitmap & $FF
        bitmap := bitmap >> 8
        chkM -= c
        data.writeLocation(eepromDev,(i<<15)+busBitmap+j,c,16,8)
        waitcnt(clkfreq / 200 + cnt)
      data.writeLocation(eepromDev,(i<<15)+checkMap,chkM,16,8)
      waitcnt(clkfreq / 200 + cnt)
      return

PUB markNewName(namePtr,addr,opt) | pin
'' This routine initializes the code space at the specified address
'' on the I2C bus with the supplied file name.  It does check that
'' the code space is either marked as deleted or is uninitialized.
'' The sign bits of the name are used for 12 option bits given here.
'' The address format is pin pair in bits 21..19, page number in
'' bits 18..15, and address in 14..0.
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  addr &= $7FF000                 ' If option bits present, mask them
  if checkSums(0,addr) <> true
    writeFilename(namePtr,addr,opt)
    return false
  abort -string("file not created, space not valid")

PUB markDeleted(namePtr,addr) | pin
'' This routine marks as deleted the code space at the specified
'' address on the I2C bus.  It does check that the code space
'' has the specified name unless the pointer is zero.
'' The sign bits of the name are used for 12 option bits, zeroed here.
'' The address format is pin pair in bits 21..19, page number in
'' bits 18..15, and address in 14..0.
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  addr &= $7FF000                 ' If option bits present, mask them
  if checkSums(namePtr,addr) <> true
    writeFilename(0,addr,0)
    return false
  abort -string("file not deleted, space not valid")

PUB searchFile(namePtr) | i, j, b, addr, t, k
'' This routine searches all available EEPROMs for an object code
'' space with the name given.  It returns the pin pair in bits 21-19,
'' the 32K page number in bits 18-15, and the address in bits 14-0.
'' If the code space is not found, an abort is done with the
'' negative address of an error message returned.  If namePtr is
'' zero, this looks for an uninitialized or deleted code space.
'' The sign bits of the name are used for 12 option bits.  These
'' are returned in the least significant 12 bits of the address.
'' The name "BOOT" is a special case for the boot EEPROM.
  if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
    return ldr#i2cBootSCL << 18
  repeat i from 0 to ldr#maxPinPairs-2 step 2   ' Pin pairs
    b := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & b) <> 0          ' Is this one present?
        data.init(i+1,i)         ' Use pin pair
        addr := (i << 18) | (j << 15)  ' Check name,chksums        
        if (t := checkSums(namePtr,addr)) <> true
          return addr | t        ' Match, return location
  abort -string("file not found") ' and option bits

PRI checkSums(namePtr, addr) | i,p,c,eq,chkS,chkM,devA,opt,xxx
'' This routine checks the filename and bitmap checksums.
'' If those are correct, it compares the filename to the
'' one supplied (or to all blanks if none supplied) and
'' returns the options bits if all match.  If the checksums
'' fail and no name was supplied, a false is returned (to
'' catch the uninitialized situation).  Otherwise return true.
'' The boot EEPROM will never be found.  It's a special case.
'' In particular, the name "BOOT" is always the boot EEPROM.
  result := 0
  if ((addr >> 18) & $1E) == ldr#i2cBootSCL
    return true                   ' boot EEPROM never found
  if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
    return 0                      ' name BOOT is always boot EEPROM
  devA := eepromDev | ((addr >> 15) & %1110)
  addr &= $F000                   ' If option bits present, mask them
  chkM := data.readLocation(devA,addr+checkMap,16,8)
  repeat i from 0 to 1
    chkM += data.readLocation(devA,addr+busBitmap+i,16,8)
  if (p := namePtr) == 0
    p := string("            ")
  eq := true
  chkS := data.readLocation(devA,addr+checkSum,16,8)
  repeat i from 0 to 11
    c := data.readLocation(devA,addr+filename+i,16,8)
    chkS += c
    eq := eq and ((c & $7F) == byte[p++])
    result |= (c >> 7) << i
  if ((chkS & $FF) == $FF) and ((chkM & $FF) == $FF)
    if not eq
      result := true
  else
    result := namePtr <> 0

PRI writeFilename(namePtr, addr, opt) | i, p, c, chkS, chkM, m, devA, pin
'' This routine writes the filename and bitmap fields and their checksums
'' to the end of the 32Kx8 space given by "addr".  The option information
'' in the sign bits of the characters is provided in "opt".
'' The name "BOOT" is a special case for the boot EEPROM and is not stored.
  m := getBusBitmap             ' 
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  devA := eepromDev | ((addr >> 15) & %1110)
  addr &= $F000                   ' If option bits present, mask them
  chkM := $FF
  repeat i from 0 to 1            ' Write bitmap bytes & compute checksum
    c := m & $FF
    m := m >> 8
    chkM -= c
    data.writeLocation(devA,addr+busBitmap+i,c,16,8)
    waitcnt(clkfreq / 200 + cnt)
  data.writeLocation(devA,addr+checkMap,chkM,16,8)
  waitcnt(clkfreq / 200 + cnt)    ' Write bitmap checksum
  if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
    return                        ' Only store the bitmap if boot EEPROM
  if (p := namePtr) == 0
    p := string("            ")   ' If no name, use blanks
  chkS := $FF
  repeat i from 0 to 11
    c := byte[p++] & $7F | (opt << 7 & $80)
    opt >>= 1                     ' Insert option bits,
    chkS -= c                     '  compute checksum,
    data.writeLocation(devA,addr+filename+i,c,16,8)
    waitcnt(clkfreq / 200 + cnt)  '   and write name bytes
  data.writeLocation(devA,addr+checkSum,chkS,16,8)
  waitcnt(clkfreq / 200 + cnt)    ' Write name checksum

PUB probeEEPROMs(SCLpin) : bitmap | device, devA, addr, b0, b0a, b1, b1a, val, seed, t
'' This routine first attempts to see if an EEPROM is present.
'' A series of 32Kx8 address spaces could be provided by one
'' device or several.  This routine builds a bitmap showing which
'' have been provided if any and returns a value suitable for
'' placing in the system's EEPROM table.
  bitmap := 0
  seed := long[ldr#systemPtr] + ldr#sys_random
  data.init(SCLpin+1,SCLpin)
  if data.testBus == data#_none       ' Is there at least a pullup on SDA?
    repeat device from 0 to 14 step 2  ' If so, go through the EEPROM devices
      devA := eepromDev | device
      if data.devicePresent(devA)     ' Add the ones that respond to bitmap
        bitmap |= |<device            ' Now we have to see what sizes they are
        addr := ?long[seed] & $7FFF   ' Generate random 32K address
        b0 := data.readLocation(devA,addr,16,8)        ' Read current data there
        b1 := data.readLocation(devA,addr|$8000,16,8)  '  in each 32K page
        repeat                        ' Generate a random byte not in either
          val := ?long[seed] & $FF    '  of the two 32K pages.  They might be
        until val <> b0 and val <> b1  '  the same physical location.
        data.writeLocation(devA,addr|$8000,val,16,8)  ' Write byte in 2nd 32K page
        waitcnt(clkfreq / 200 + cnt)  ' (5ms for write)
        b0a := data.readLocation(devA,addr,16,8)      ' Now see which 32K page changed
        b1a := data.readLocation(devA,addr|$8000,16,8)
        if b0 == b0a and b1a == val   ' If the change is in the second 32K only
          bitmap |= |<(device + 1)    '  then we really have memory there
          data.writeLocation(devA,addr|$8000,b1,16,8)  ' Restore the original data
        else
          data.writeLocation(devA,addr,b0,16,8)  ' If not there, restore first 32K
        waitcnt(clkfreq / 200 + cnt)  ' (5ms for write)

PUB dumpMemory(pin,addr,size) | i, c, p, page, first, a
'' This routine dumps a portion of the RAM or EEPROM to the display.
'' The format is 16 bytes wide with hexadecimal and printable characters.
'' If no EEPROM is present at that location, an error message is displayed.
  page := (addr >> 15) & $1F
  addr &= $7FFF
  if pin <> -1
    data.init(pin+1,pin)
    if data.testBus <> data#_none
      vga.str(string("*** no I2C bus on those pins",vga#Cr))
      return
    if not data.devicePresent(page&%1110|eepromDev)
      vga.str(string("*** no EEPROM at that address",vga#Cr))
      return
  if addr + size > $7FFF               ' truncate to end of code space
    size := $8000 - addr
  first := true
  p := addr & $7FF0
  repeat while p < (addr + size)
    vga.str(string("    "))
    if pin < (ldr#maxPinPairs << 1)
      vga.decFixed(pin,2)
      vga.out(":")
    else
      vga.str(string("   "))
    if first
      vga.hex(addr,5)
      first := false
    else
      vga.hex(p,5)
    vga.out3(vga#MoveX,62,"|")
    repeat i from 0 to 15
      if p => addr and p < (addr + size)
        if pin < (ldr#maxPinPairs << 1)
          c := data.readLocation((page&%1110)|eepromDev,(page&%0001)<<15|p,16,8)
        else
          c := byte[p]
        vga.out2(vga#MoveX,3*i+13)
        vga.hex(c,2)
        vga.out2(vga#MoveX,i+63)
        if c < " " or c > "~"
          vga.out(".")
        else
          vga.out(c)
      p++
    vga.out4(vga#MoveX,79,"|",vga#Cr)
  
PUB fileList | i,j,k,bitmap,devA,addr,chkM,chkS,t,opt,a,b,c
'' This routine displays a list of all files currently known to the system.
  repeat i from 0 to 15          ' Pin pairs
    bitmap := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & bitmap) <> 0     ' Is this one present?
        data.init(i<<1+1,i<<1)   ' Use pin pair
        devA := eepromDev | (j & %1110)  ' Device address
        addr := (j & 1) << 15    ' 16 bit byte address
        chkM := data.readLocation(devA,addr+checkMap,16,8)
        repeat k from 0 to 1     ' Make sure bitmap checksum correct
          chkM += data.readLocation(devA,addr+busBitmap+k,16,8)
        chkS := data.readLocation(devA,addr+checkSum,16,8)
        opt := 0                 ' Read filename, extract option bits,
        repeat k from 0 to 11    '  and make sure checksum is correct
          t := data.readLocation(devA,addr+filename+k,16,8)
          chkS += t
          byte[@a][k] := t & $7F
          opt |= t >> 7 << k
        vga.str(string("    "))
        vga.decFixed(i<<1,2)
        vga.out(":")
        vga.hex(j<<15,5)
        if (chkS & $FF) == $FF and (chkM & $FF) == $FF
          if a == $20202020 and b == $20202020 and c == $20202020
            vga.str(string(" Unused (deleted)",vga#Cr))
          else
            vga.str(string(" <"))
            repeat k from 0 to 11
              vga.out(byte[@a][k])
            vga.str(string("> [$"))
            vga.hex(opt,3)
            if opt & opt_text
              vga.out(",")
              vga.str(string("text"))
            vga.str(string("]",vga#Cr))
        else
          if (i == 14) and (j == 0)
            vga.str(string(" Boot EEPROM",vga#Cr))
          else
            vga.str(string(" Unused (uninitialized)",vga#Cr))

PUB readEEPROM(addr,bytes) | pin, devA, i
'' This routine returns up to 4 bytes (LSB to MSB) from the specified location in an EEPROM.
'' The address format is pin pair in bits 21..19, page number in bits 18..15, address in 14..0.
  pin := addr >> 18 & $1E
  data.init(pin+1,pin)
  devA := %10100000 | ((addr >> 15) & %1110)
  addr &= $FFFF
  repeat i from 0 to (bytes <# 4) - 1
    result |= data.readLocation(devA,addr+i,16,8) << (i << 3)

PUB copyEEPROM(addrTo,addrFrom,size) | i,w,r,x,b0,b1,b2,b3,b4,b5,b6,b7
'' This routine copies "size" bytes from one EEPROM address to another.
'' Both addresses MUST be on a 32 byte boundary because of paged write
'' wraparound issues.  If the entire range of bytes is within a 32 byte
'' boundary, this does not apply.  If it is known that the EEPROM being
'' written has a larger page size, then the larger page size applies.
'' "size" is limited to 32768 bytes, but, because the EEPROMs are
'' readdressed every 32 bytes, the address range can extend over a 32K
'' boundary (if there is EEPROM there).
  w := size & $7FE0
  x := vga.windowCol
  repeat i from 0 to w-32 step 32  ' Process full blocks of 32
    vga.out2(vga#MoveX,x)
    vga.decFixed(size-i,5)
    readDataSequential(addrFrom+i,@b0,32)
    writeDataPaged(addrTo+i,@b0,32)
  if (r := size & $1F) > 0
    vga.out2(vga#MoveX,x)
    vga.decFixed(r,5)
    readDataSequential(addrFrom+w,@b0,r)
    writeDataPaged(addrTo+w,@b0,r)
  vga.out2(vga#MoveX,x)

PRI readDataSequential(address, buffAdr, count) | i, pin, ackbit, devAddr
  ' This reads "count" bytes into a buffer at "buffAdr" beginning at location "address"
  ' in an EEPROM that supports sequential reads.  This returns a sequence of acknowledge
  ' bits lsb last (up to 32 with the msb being "sticky") with NAK being a one bit.
  ' Only the 4 device selection and addressing write ACK/NAKs are present on reads.
  pin := address >> 18 & $1E
  data.init(pin+1,pin)
  devAddr := (address >> 15) & %1110 | eepromDev
  ackbit := data#_i2cACK
  if data.isStarted == true  
    data.i2cStart
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(devAddr | 0, 8)
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(address >> 8 & $FF, 8)
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(address & $FF, 8)         
    data.i2cStart
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(devAddr | 1, 8)
    if count > 1
      repeat i from 0 to count-2
        byte[buffAdr+i] := data.i2cRead(8,data#_i2cACK)
    if count > 0
      byte[buffAdr+count-1] := data.i2cRead(8,data#_i2cNAK)
    data.i2cStop
  else
    ackbit := data#_i2cNAK
  return ackbit

PRI writeDataPaged(address, buffAdr, count) | i, pin, ackbit, devAddr
  ' This writes "count" bytes from a buffer at "buffAdr" beginning at location "address"
  ' in an EEPROM that supports paged writes.  This returns a sequence of acknowledge
  ' bits lsb last (up to 32 with the msb being "sticky") with NAK being a one bit.
  ' The first 4 ACK/NAK bits are for the device selection and addressing writes.
  ' As is usual with paged writes, the addresses wrap around to the beginning of a page.
  ' The page size is device dependent and is usually device size dependent.  For 32K x 8,
  ' this is 64.  For 64K x 8, this is 128.  For 128K x 8, this is 256.
  pin := address >> 18 & $1E
  data.init(pin+1,pin)
  devAddr := (address >> 15) & %1110 | eepromDev
  ackbit := data#_i2cACK      
  if data.isStarted == true
    data.i2cstart
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(devAddr | 0,8)
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(address >> 8 & $FF, 8)
    ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(address & $FF, 8)
    if count > 1
      repeat i from 0 to count-2
        ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(byte[buffAdr+i],8)
    if count > 0
      ackbit := (ackbit & $80000000) | (ackbit << 1) | data.i2cWrite(byte[buffAdr+count-1],8)
    data.i2cStop
  else
    ackbit := data#_i2cNAK
  return ackbit