'' This object provides an EEPROM loader that takes a system EEPROM address (pin pair # in
'' bits 22..19, 32Kx8 Page # in bits 18..15, 32Kx8 address offset in bits 14..0), loads
'' the program located there into RAM and starts the SPIN interpreter in the current COG.
'' The "vbase" and "dbase" pointers in the program are used to determine how much should
'' be loaded from the EEPROM.  EEPROMs used should be rated for bus speeds of 400KHz min.
'' and the loader assumes an 80MHz system clock (configured with constant declarations).

'' A routine "checkPresence" checks for the presence of EEPROM at the system EEPROM address
'' provided by attempting to select the EEPROM at that I/O pin pair and device select code.
'' This routine cannot distinguish between a 32Kx8 EEPROM and a 64Kx8 EEPROM since the same
'' 32K of data appears in both halves of the 16 bit address space and, unlike the routines
'' in the "Operating System", these do not attempt to write to one half of the address space
'' and check to see where the data actually appears (then restores the location written).

'' This object also provides constant declarations for sizes of various system tables and
'' buffers and offsets to their pointers in "high memory".  The offsets within the system
'' work area are declared here as well.  Some simple methods are provided for memory allocation
'' and for registering and unregistering COGs as being persistant.  Any COGs not registered
'' will be stopped when a program is read into RAM using the RUN command.  These routines
'' depend on the global area being initialized by the "Operating System" initialization
'' routines and these should not be called unless the global area has been initialized since
'' the last reset.  The loader and the "checkPresence" routines are not dependent on the
'' global area and can be called by any program that includes this object.

CON
  maxPinPairs = 16                             ' Maximum number of possible I2C busses
  i2cBootSCL  = 28                             ' I2C Clock pin for boot EEPROM
  i2cBootSDA  = 29                             ' I2C Data pin for boot EEPROM

'' System work area definitions

  sys_cog     = 0                              ' Contains bitmap for cogs used by system
  sys_random  = 4                              ' Random seed for "?" operator
  sys_eeproms = 8                              ' Table of EEPROMs, one for each I/O pair
                                               ' Each word entry contains a bitmap of the
                                               ' 32Kx8 address blocks that are present on
                                               ' the I2C bus on the corresponding pair of
                                               ' I/O pins (if any).  Size=maxPinPairs*2

'' System work areas, buffer, and table sizes

  systemSize  = 40                             ' # of bytes for system information
  videoSize   = 24 + (16 * 8)                  ' # of bytes for video
  screenSize  = 128 * 64                       ' # of bytes for screen buffer
  colorSize   = 64 * 2                         ' # of bytes for color buffer
  cursorSize  = 8                              ' # of bytes for screen cursor info
  keybdSize   = 80                             ' # of bytes needed for keyboard
  videoNeeded = videoSize + screenSize + colorSize + cursorSize
  totalNeeded = systemSize + videoNeeded + keybdSize

'' The following are fixed locations assigned at the top of RAM memory to avoid
'' having them overwritten when a new program is loaded.  Most of these are pointers
'' to work areas allocated "below" them.  The last pointer is reserved for non-system
'' programs

  endMemory   = $8000                          ' Address past the end of RAM
  initMarker  = endMemory -  4                 ' Unique marker for initialized data
  memPtr      = endMemory -  8                 ' Address of last allocated long
  systemPtr   = endMemory - 12                 ' Address of system information
  videoPtr    = endMemory - 16                 ' Address of video work area
  screenPtr   = endMemory - 20                 ' Address of screen buffer
  colorPtr    = endMemory - 24                 ' Address of color buffer
  cursorPtr   = endMemory - 28                 ' Address of screen cursor info
  keybdPtr    = endMemory - 32                 ' Address of keyboard work area
  userPtr     = endMemory - 36                 ' Address of work area for "user"
  endFree     = userPtr                        ' Address past initial free memory

'' Internal SPIN information at fixed locations in RAM/EEPROM image

  spin_pbase  = $0006                          ' must be $0010
  spin_vbase  = $0008                          ' number of longs loaded times 4
  spin_dbase  = $000A                          ' above where $FFF9FFFF's get placed
  spin_pcurr  = $000C                          ' points to SPIN code
  spin_dcurr  = $000E                          ' points to local stack

PUB bootFromEEPROM(addr)
  coginit(cogid,@loadHubMemory,(addr >> 7) & $FFFC)    ' Run assembly loader in this COG

PUB allocate(size)                                     ' Allocate block of size bytes
  return long[memPtr] := (long[memPtr] - size) & !3    ' Always a long address

PUB deallocate                                         ' User work area no longer in use
  long[memPtr] := long[keybdPtr]

PUB registerCOG(bitmap)                                ' Register "persistant" COGs
  long[long[systemPtr]+sys_cog] |= bitmap

PUB unregisterCOG(bitmap)                              ' Mark COGs no longer "persistant"
  long[long[systemPtr]+sys_cog] &= !bitmap

PUB checkPresence(addr) | SCLpin, SDApin, devA, driven
'' This routine checks to be sure there is an I2C bus and an EEPROM at the
'' specified address.  This is a minimal special version of this function
'' in the OS_i2cObject and OS_fileSystem objects since this may be the only
'' SPIN I2C access function required (prior to attempting to boot or copy
'' the boot EEPROM.
  devA := ((addr >> 15) & %1110) | %10100000
  SCLpin := (addr >> 18) & $1E
  SDApin := SCLpin + 1
  driven := SCLpin == i2cBootSCL
  dira[SDApin]~                                 ' leave I2C bus floating now
  dira[SCLpin]~
  outa[SDApin]~                                 ' prepare to drive active low
  outa[SCLpin]~
  if ina[SDApin] <> 1                           ' check for pullup(s)
    return false
  if not driven and (ina[SCLpin] <> 1)
    return false
  if driven                                     ' start condition
    outa[SCLpin]~~
    dira[SCLpin]~~
  else
    dira[SCLpin]~
  dira[SDApin]~~
  outa[SDApin]~
  if not driven
    repeat until ina[SCLpin] == 1              ' end of start condition
  dira[SDApin] ~~                              ' set both line to outputs
  dira[SCLpin] ~~                              '  in preparation for writing
  outa[SCLpin]~                                ' set the clock low
  repeat 8                                     ' send the 8 bits, msb first
    outa[SDApin] := devA >> 7                  '  set the SDA while SCL low
    outa[SCLpin]~~                             '  toggle SCL high
    outa[SCLpin]~                              '  toggle SCL low
    devA <<= 1                                 '  next lsb
  dira[SDApin]~                                ' setup for ACK, change to input
  outa[SCLpin]~~                               ' toggle SCL high
  result := ina[SDApin] == 0                   ' read in the ACK
  outa[SCLpin]~                                ' toggle SCL low
  dira[SDApin]~~                               ' leave SDA low
  outa[SDApin]~      
  if driven                                    ' stop condition
    outa[SCLpin]~~
    dira[SCLpin]~~
  else
    dira[SCLpin]~
  dira[SDApin]~                                ' end of stop condition

CON                                            ' These constants all assume an 80MHz clock
  nsPerInst   =   50                           ' ns per basic instruction
  i2cSetHold  =  600 / nsPerInst               ' instructions per minimum setup/hold times
  i2cClkHigh  =  600 / nsPerInst               ' instructions per minimum clock high time
  i2cClkLow   = 1300 / nsPerInst               ' instructions per minimum clock low time

DAT
                        org     0
loadHubMemory           mov     i2cAddr,PAR
                        and     i2cAddr,i2cAddrMask     ' Isolate and position EEPROM address bits
                        shl     i2cAddr,#7
                        mov     bootTemp,PAR            ' Isolate the SCL pin number
                        shr     bootTemp,#11
                        and     bootTemp,#%11110        ' Generate bit mask for SCL
                        mov     i2cSCL,#1
                        shl     i2cSCL,bootTemp
                        mov     i2cSDA,i2cSCL           ' Generate bit mask for SDA from SCL
                        shl     i2cSDA,#1
                        mov     i2cBufAdr,#0            ' Start at RAM location 0
                        mov     i2cCount,#$10
                        call    #readLocation           ' Get preamble first (with length)
                if_nc   jmp     #:loadHubError
                        rdword  i2cCount,#spin_vbase    ' Get byte count from the loaded program
                        sub     i2cCount,#$10
                        call    #readLocation           ' Now get the rest of the program
                if_nc   jmp     #:loadHubError
                        rdword  bootTemp,#spin_dbase    ' Only the code and variable space is
                        sub     bootTemp,#4             '  loaded, put in the stack marker
                        wrlong  stackMark,bootTemp      '   just below the start of the stack
                        sub     i2cTemp,#4
                        wrlong  stackMark,bootTemp
                        rdbyte  bootTemp,#$0004         ' If xtal/pll enabled, start up now
                        and     bootTemp,#$F8           '  while remaining in current mode
                        clkset  bootTemp
:delay                  djnz    time_xtal,#:delay       ' Allow 20ms@20MHz for xtal/pll to settle
                        rdbyte  bootTemp,#$0004         ' Switch to selected clock
                        clkset  bootTemp
                        coginit interpreter             ' Reboot this cog with interpreter

:loadHubError           cogid   bootTemp                ' If an error occurs, stop the cog
                        cogstop bootTemp

i2cAddrMask             long    $0FFC                   ' Isolate 32k page and offset
i2cBootSCLm             long    $10000000               ' Boot EEPROM clock line = pin 28
stackMark               long    $FFF9FFFF
time_xtal               long    20 * 20000 / 4 / 1      ' 20ms (@20MHz, 1 inst/loop)
interpreter             long    $0001 << 18 + $3C01 << 4 + %0000
bootTemp                long    0

'' Mid level I2C routines.

'' Read a sequence of bytes from an attached I2C EEPROM into a buffer in hub memory.
'' The starting I2C address is provided as well as a byte count and starting cog address.
'' The caller sets up the masks for the SCL and SDA pins and the timing information for
'' the bus speed required.  EEPROM addresses are assumed for Atmel or Microchip or
'' equivalent devices in the 32Kx8 to 128Kx8 range with the high order 3 bits in the
'' device select byte and the low order 16 bits in the device register bytes (MSB,LSB).
'' Exit with (c) set if a NAK was received when not expected.  The write routine works
'' similarly except that it allows for paged writes.  Since all the devices in the range
'' supported (32Kx8 to 128Kx8) have pages of at least 64 bytes, this is the size supported.
'' When the address crosses a page boundary, a Stop Sequence is forced, then the device is
'' readdressed about 5ms later to allow for the write cycle to complete.  For sequential reads,
'' the smallest device is assumed to be 32Kx8 and, when reads cross a 32K boundary, a Stop
'' Sequence is forced and the device is readdressed.  The page mask (i2cPageMask) can be
'' changed by the caller if other page sizes (for paged writes) are to be supported.  It can
'' be set to all one bits for devices like Ramtron RAM that have no write delays.

readCheck               test    i2cAddr,i2c32kPage wz   ' Is this the last byte in a
                if_z    jmp     #readNext               '  32K page?  If so, force Stop
                        call    #i2cStop                '   Sequence, and start over
readLocation            call    #i2cSendAddr            ' Address the device
                if_c    jmp     #readDoStop             '  as if for a write
                        mov     i2cData,i2cDevA
                        or      i2cData,#1
                        call    #i2cStart               ' Reselect the device
                        call    #i2cWrite               '  for read cycles
                if_c    jmp     #readDoStop
readNext                mov     i2cData,#0
                        mov     i2cMask,%10000000
                        mov     i2cTemp,i2cAddr         ' Is this the last byte in a
                        and     i2cTemp,i2c32kPage      '  32K page?  If so, force a
                        cmp     i2cTemp,i2c32kPage wz   '   NAK with this byte
                        muxz    i2cZero,#1      nr,wc
                if_nc   cmp     i2cCount,#1        wz   ' Is this the last byte to read?
                if_nc   muxz    i2cZero,#1      nr,wc   ' If so, force a NAK with this byte
                        call    #i2cRead
                        wrbyte  i2cData,i2cBufAdr       ' Write the byte to hub memory
                        add     i2cBufAdr,#1
                        add     i2cAddr,#1              ' Increment the addresses and
                        djnz    i2cCount,#readCheck     '  decrement the byte count
readDoStop              call    #i2cStop                ' Do a Stop Sequence if done
readLocation_ret        ret

writeCheck              test    i2cAddr,i2cPageMask wz  ' Did we roll over the end of a page?
                if_z    jmp     #writeNext              ' If not, continue with the next byte
                        call    #i2cStop                ' If so, generate a Stop Sequence
:writeWait              mov     i2cData,i2cDevA
                        call    #i2cStart               ' Now do acknowledge polling ...
                        call    #i2cWrite               ' Try to setup a write.  If not
                if_c    jmp     #:writeWait             '   acknowledged, keep trying.
                        call    #i2cStop
writeLocation           call    #i2cSendAddr            ' Address the device
                if_c    jmp     #writeDoStop
writeNext               mov     i2cMask,%10000000
                        rdbyte  i2cData,i2cBufAdr       ' Read the byte from hub memory
                        call    #i2cWrite               '  and transfer to the EEPROM
                if_c    jmp     #writeDoStop
                        add     i2cBufAdr,#1            ' Increment the addresses and
                        add     i2cAddr,#1              '  decrement the byte count
                        djnz    i2cCount,writeCheck
                        call    #i2cStop                ' Do a Stop Sequence if done
                        add     i2cPageMask,#1  nr,wz   ' Check for absence of write delay
                if_z    jmp     #writeLocation_ret      '  (as for Ramtron RAM)
:writeWait              mov     i2cData,i2cDevA
                        call    #i2cStart               ' Now do acknowledge polling ...
                        call    #i2cWrite               ' Try to setup a write.  If not
                if_c    jmp     #:writeWait             '   acknowledged, keep trying.
writeDoStop             call    #i2cStop
writeLocation_ret       ret

'' Set up a 3 byte sequence with device select code, then 16 bit device address (MSB..LSB).
'' Do a start sequence, then send the 3 bytes.  Set C flag if any NAKs occurred.

i2cSendAddr             call    #debugInit              ' Display starting at row 0 col 50
                        add     debugDsp,#50
                        mov     debugVal,i2cAddr        ' Show current address
                        call    #debugHexLong
                        mov     debugChr,#" "
                        call    #debugOut
                        mov     debugVal,i2cSCL
                        or      debugVal,i2cSDA
                        call    #debugHexLong
                        mov     debugChr,#"|"
                        call    #debugOut
                        mov     i2cDevA,i2cAddr
                        and     i2cDevA,#$FF            ' Bits 7-0 of address sent as last byte
                        shl     i2cDevA,#16
                        mov     i2cData,i2cAddr
                        and     i2cData,i2cHighAddr     ' Bits 15-8 of address sent as first byte
                        or      i2cData,i2cDevA
                        mov     i2cDevA,i2cAddr
                        shr     i2cDevA,#15             ' Bits 18-16 of address in device address
                        and     i2cDevA,#%00001110
                        or      i2cDevA,#%10100000
                        or      i2cData,i2cDevA
                        call    #i2cStart               ' Generate start condition
                        call    #i2cWrite               ' Send 1st byte - device address
                if_nc   call    #i2cWrite               ' Send 2nd byte - MSB of 16 bit address
                if_nc   call    #i2cWrite               ' Send 3rd byte - LSB of 16 bit address
i2cSendAddr_ret         ret

i2cPageMask             long    %111111                 ' Page address boundary (for paged writes)
i2cHighAddr             long    $FF00                   ' High order address byte
i2c32kPage              long    $7FFF                   ' 32K address boundary
i2cZero                 long    0                       ' A zero long value
i2cAddr                 long    0                       ' Starting address in EEPROM
i2cCount                long    0                       ' Number of bytes to read / write
i2cBufAdr               long    0                       ' Buffer address in cog memory
i2cDevA                 long    0                       ' Save device address
i2cTemp                 long    0

'' Low level I2C routines.  These are designed to work either with a standard I2C bus
'' (with pullups on both SCL and SDA) or the Propellor Demo Board (with a pullup only
'' on SDA).  Timing can be set by the caller to 100KHz, 400KHz, or 1MHz.

'' Do I2C Start Sequence.  This assumes that both SCL and SDA are floating (inputs).
'' The start sequence is where SDA goes from HIGH to LOW while SCL is HIGH.

i2cStart                andn    outa,i2cSDA             ' Prepare to drive SDA low
                        or      outa,i2cSCL
                        or      dira,i2cSCL             ' Active drive SCL high
                        or      dira,i2cSDA             '  active drive SDA low
                        mov     i2cTime,#i2cSetHold
:wait1                  djnz    i2cTime,#:wait1         ' Allow for minimum setup time
                        andn    outa,i2cSCL             '  with SCL low and SDA low
i2cStart_ret            ret                             

'' Do I2C Stop Sequence.  This assumes that SCL is low and SDA is indeterminant.
'' The stop sequence is where SDA goes from LOW to HIGH while SCL is HIGH.
'' i2cStart must have been called prior to calling this routine for initialization.

i2cStop                 or      dira,i2cSDA             ' Active drive SDA low
                        or      outa,i2cSCL             ' Active drive SCL high
                        or      dira,i2cSCL
                        mov     i2cTime,#i2cSetHold
:wait1                  djnz    i2cTime,#:wait1         ' Wait for hold time and
                        andn    dira,i2cSDA             '  let SDA float to high
                        mov     i2cTime,#i2cSetHold
:wait2                  djnz    i2cTime,#:wait2         ' Wait for hold time and
                        andn    dira,i2cSCL             '  let SCL float to high
                        mov     i2cTime,#i2cClkLow
:wait3                  djnz    i2cTime,#:wait3         ' Wait for minimum bus free
                        test    i2cZero,#0      nr,wc   '  time and force (c) false
i2cStop_ret             ret                             '   which implies ACK

'' Write I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant, and i2cTime is the time to the next SCL low->high transition.
'' The (c) flag will be set on exit from ACK/NAK with ACK == low and NAK == high.
'' Bytes are handled in "little-endian" order so these routines can be used with
'' words or longs although the bits are in msb..lsb order.

i2cWrite                mov     i2cBitCnt,#8
                        mov     debugVal,i2cData
                        call    #debugHexByte
:i2cWriteBit            mov     i2cTime,#i2cClkLow
:wait1                  djnz    i2cTime,#:wait1         ' Wait for minimum SCL low
                        or      outa,i2cSCL             '  and drive SCL high
                        test    i2cData,i2cMask wz
              if_z      or      dira,i2cSDA             ' Copy data bit to SDA
              if_nz     andn    dira,i2cSDA
                        mov     i2cTime,#i2cClkHigh
:wait2                  djnz    i2cTime,#:wait2         ' Wait for minimum SCL high
                        andn    outa,i2cSCL             '  and drive SCL low
                        ror     i2cMask,#1              ' Go do next bit if not done
                        djnz    i2cBitCnt,#:i2cWriteBit
                        andn    dira,i2cSDA             ' Switch SDA to input and
                        mov     i2cTime,#i2cClkLow      '  wait for minimum SCL low
:wait3                  djnz    i2cTime,#:wait3
                        or      dira,i2cSCL             ' Take SCL high and wait for
                        mov     i2cTime,#i2cClkHigh     '  minimum high time
:wait4                  djnz    i2cTime,#:wait4
                        test    i2cSDA,ina         wc   ' Sample SDA (ACK/NAK) and
                        andn    outa,i2cSCL             '  take SCL low
                        rol     i2cData,#16             ' Prepare for multibyte write
i2cWrite_ret            ret

'' Read I2C data.  This assumes that i2cStart has been called and that SCL is low,
'' SDA is indeterminant, and i2cTime is the time to the next SCL low->high transition.
'' ACK/NAK will be copied from the (c) flag on entry with ACK == low and NAK == high.
'' Bytes are handled in "little-endian" order so these routines can be used with
'' words or longs although the bits are in msb..lsb order.

i2cRead                 mov     i2cBitCnt,#8
:i2cReadBit             mov     i2cTime,#i2cClkLow      ' Wait for minimum SCL low
:wait1                  djnz    i2cTime,#:wait1
                        or      outa,i2cSCL             ' Drive SCL high, and make
                        andn    dira,i2cSDA             '  sure SDA is set to input
                        mov     i2cTime,#i2cClkHigh
:wait2                  djnz    i2cTime,#:wait2         ' Wait for minimum SCL high,
                        test    i2cSDA,ina         wz   '  sample SDA just before
                        andn    outa,i2cSCL             '   taking SCL low
              if_nz     or      i2cData,i2cMask         ' Accumulate data bits
              if_z      andn    i2cData,i2cMask
                        ror     i2cMask,#1              ' Shift the bit mask and
                        djnz    i2cBitCnt,#:i2cReadBit  '  continue until done
                        mov     i2cTime,#i2cClkLow
:wait3                  djnz    i2cTime,#:wait3         ' Wait for end of SCL low
                        or      outa,i2cSCL             '  and take SCL high
              if_nc     andn    dira,i2cSDA             ' Copy the ACK/NAK bit to SDA
              if_c      or      dira,i2cSDA
                        mov     i2cTime,#i2cClkHigh
:wait4                  djnz    i2cTime,#:wait4       ' Wait for minimum SCL high
                        andn    outa,i2cSCL             '  and take SCL low
                        rol     i2cMask,#16             ' Prepare for multibyte read
i2cRead_ret             ret

'' Local variables for low level I2C routines

i2cSCL                  long    0                       ' Bit mask for SCL
i2cSDA                  long    0                       ' Bit mask for SDA
i2cTime                 long    0                       ' Used for timekeeping
i2cData                 long    0                       ' Data to be transmitted / received
i2cMask                 long    0                       ' Bit mask for bit to be tx / rx
i2cBitCnt               long    0                       ' Number of bits to tx / rx

'' 32 bit unsigned division yielding quotient and remainder

divide                  mov     temp,#32                ' Divide a 32 bit unsigned dividend
                        mov     remainder,#0            '  by a 32 bit unsigned divisor to
:loop                   shl     dividend,#1        wc   '  get a 32 bit unsigned quotient
                        rcl     remainder,#1            '  and a 32 bit unsigned remainder
                        cmpsub  remainder,divisor  wc,wr
                        rcl     quotient,#1
                        djnz    temp,#:loop
divide_ret              ret

temp                    long    0
dividend                long    0
remainder               long    0
divisor                 long    0
quotient                long    0

'' Debugging routines that write directly to the display buffer

debugHexPart            mov     debugShift,#0
                        mov     debugChr,#0
:debugHexLoop           rol     debugVal,#4
                        mov     debugDig,debugVal
                        and     debugDig,#%1111
                        add     debugDig,#"0"
                        cmp     debugDig,#"9"+1       wc
              if_nc     add     debugDig,#"A"-("9"+1)
                        shl     debugDig,debugShift
                        or      debugChr,debugDig
                        add     debugShift,#8
                        djnz    debugCnt,#:debugHexLoop
                        call    #debugOut
debugHexPart_ret        ret

debugHexByte            mov     debugCnt,#2
                        shl     debugVal,#24
                        call    #debugHexPart
debugHexByte_ret        ret

debugHexWord            mov     debugCnt,#4
                        shl     debugVal,#16
                        call    #debugHexPart
debugHexWord_ret        ret

debugHexLong            mov     debugHexPart_ret,debugHexLong_ret
                        mov     debugCnt,#4
                        call    #debugHexPart
                        mov     debugCnt,#4
                        call    #debugHexPart
debugHexLong_ret        ret

debugInit               rdword  debugDsp,screenPtrAdr    ' Write directly to the screen
debugInit_ret           ret                              '  initially row 0, column 0

debugOutNext            wrbyte  debugChr,debugDsp        ' Write char & move to next
                        add     debugDsp,#1             
                        shr     debugChr,#8              ' Order is LSB .. MSB          
debugOut                or      debugChr,#0        nr,wz ' Copy up to 4 chars to display
              if_nz     jmp     #debugOutNext
debugOut_ret            ret

screenPtrAdr            long    screenPtr                ' Address of screen pointer in HUB RAM
debugDsp                long    0                        ' Pointer to display memory
debugChr                long    0                        ' Up to 4 characters to output
debugVal                long    0
debugDig                long    0
debugShift              long    0
debugCnt                long    0

                          